"""
This script is a the samcut project, designed to process paired-end FASTQ files by fragmenting DNA sequences at specified restriction enzyme sites.

Copyright © 2024 Samir Bertache

SPDX-License-Identifier: AGPL-3.0-or-later

===============================================================================

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import argparse
import logging
import sys
from multiprocessing import Process, Queue
from typing import List
from .Frag import process_items
from .Read import read_fastq_gzip_simultaneously_MyWay
from .Pretreatment import SearchInDataBase, Partitionning
from .WriteAndControl import OpenOutput, ManagePigzProblems, write_pairs


# Setup logging
logging.basicConfig(level=logging.INFO)


def cut(
    source_forward: str,
    source_reverse: str,
    output_forward: str,
    output_reverse: str,
    list_enzyme: List[str],
    mode,
    seed_size,
    buffer_size: int = 200,
    num_threads: int = 8,
    borderless: bool = False,
) -> None:
    """
    Main function to process sequences based on enzyme restriction sites.

    Parameters:
        source_forward (str): Input file path for forward reads.
        source_reverse (str): Input file path for reverse reads.
        output_forward (str): Output file path for processed forward reads.
        output_reverse (str): Output file path for processed reverse reads.
        list_enzyme (List[str]): List of restriction enzymes.
        mode (str): Mode of pairing fragments, "all" or "fr".
        seed_size (int): Minimum length of fragments to keep.
        buffer_size (int, optional): Size of buffer. Defaults to 200.
        num_threads (int, optional): Number of threads to use for processing. Defaults to 8.
        borderless (bool, optional): Whether to discard ligation sites (borders). Defaults to False.

    Returns:
        None
    """
    # Threads allocations :
    TWrite, TRead, TFrag = Partitionning(num_threads)

    # Take the enzyme list and make the ligation site list
    ligation_site_list = SearchInDataBase(list_enzyme, borderless)

    try:
        # Open output files for writing
        outF, outR = OpenOutput(TWrite, output_forward, output_reverse)

        # Input and Output Queues
        Input_Buffer = Queue()
        Output_buffer = Queue()

        # Process for reading fastq files
        def read_process():
            read_fastq_gzip_simultaneously_MyWay(
                source_forward, source_reverse, Input_Buffer, TRead, TFrag
            )

        # Process for processing items
        def process_process_all():
            process_items(
                Input_Buffer,
                Output_buffer,
                ligation_site_list,
                seed_size,
                buffer_size,
                mode,
                borderless,
            )

        # Process for writing pairs
        def write_process():
            write_pairs(Output_buffer, outF, outR, TFrag)

        # Read fastq files in parallel and asynchronous
        read_p = Process(target=read_process)

        # Choose mode and Create the executor and dispatch work to it
        if mode == "all":
            print("Mode ALL selected")
            process_p_list = [
                Process(target=process_process_all) for _ in range(TFrag)
            ]
        elif mode == "fr":
            print("Mode FR selected")
            process_p_list = [
                Process(target=process_process_all) for _ in range(TFrag)
            ]
        else:
            print(f"Unknown mode: {mode}")
            sys.exit(1)

        # Create asynchronous writing
        write_p = Process(target=write_process)

        # Start processes
        read_p.start()
        for p in process_p_list:
            p.start()
        write_p.start()

        # Wait for all processes to finish
        read_p.join()
        for p in process_p_list:
            p.join()
        write_p.join()

        # Close output files and streams
        ManagePigzProblems(outF, outR, output_forward, output_reverse)

    except KeyboardInterrupt:
        print("Keyboard interrupt detected. Terminating...")
        sys.exit(0)


def main_cli():
    parser = argparse.ArgumentParser(
        description="""\
Cutsite script

Features:

    Find and Utilize Restriction Enzyme Sites: Automatically identify ligation sites from provided enzyme names and generate regex patterns to locate these sites in sequences.
    Fragmentation: Split sequences at restriction enzyme sites, creating smaller fragments.
    Multi-threading: Efficiently handle large datasets by utilizing multiple threads for decompression, fragmentation, and compression.
    Custom Modes: Supports different pairing modes for sequence fragments.

Arguments:

    --source_forward (str): Input file path for forward reads. Default is ../data/source_forward.fq.gz.
    --source_reverse (str): Input file path for reverse reads. Default is ../data/source_reverse.fq.gz.
    --output_forward (str): Output file path for processed forward reads. Default is ../data/output_forward.fq.gz.
    --output_reverse (str): Output file path for processed reverse reads. Default is ../data/output_reverse.fq.gz.
    --list_enzyme (str): Comma-separated list of restriction enzymes. Default is No restriction enzyme found.
    --mode (str): Mode of pairing fragments. Options are "all" or "fr". Default is "fr" which means Forward/Reverse.
    --seed_size (int): Minimum length of fragments to keep. Default is 20.
    --num_threads (int): Number of threads to use for processing. Default is 4.
    -b, --borderless: Option to discard ligation sites (borders).
""",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-sf", "--source_forward",
        type=str,
        default="../data/source_forward.fq.gz",
        help="Input file for forward reads",
    )
    parser.add_argument(
        "-sr", "--source_reverse",
        type=str,
        default="../data/source_reverse.fq.gz",
        help="Input file for reverse reads",
    )
    parser.add_argument(
        "-of","--output_forward",
        type=str,
        default="../data/output_forward.fq.gz",
        help="Output file for forward reads",
    )
    parser.add_argument(
        "-or", "--output_reverse",
        type=str,
        default="../data/output_reverse.fq.gz",
        help="Output file for reverse reads",
    )
    parser.add_argument(
        "-le", "--list_enzyme",
        default="No restriction enzyme found",
        type=str,
        help="Restriction Enzyme(s) used separated by coma",
    )
    parser.add_argument(
        "-m", "--mode",
        type=str,
        default="fr",
        help="Mode of modification only : all or fr",
    )
    parser.add_argument(
        "-sz", "--seed_size",
        type=int,
        default="20",
        help="Minimum lengh of fragments conserved",
    )
    parser.add_argument(
        "-nt", "--num_threads", type=int, default="8", help="Number of threads to use"
    )
    parser.add_argument(
        "-b", "--borderless", action="store_true", help="Discard ligation sites (borders)"
    )

    args = parser.parse_args()

    cut(
        source_forward=args.source_forward,
        source_reverse=args.source_reverse,
        output_forward=args.output_forward,
        output_reverse=args.output_reverse,
        mode=args.mode,
        seed_size=args.seed_size,
        num_threads=args.num_threads,
        list_enzyme=args.list_enzyme,
        borderless=args.borderless,
    )



"""
This script is a the samcut project, designed to process paired-end FASTQ files by fragmenting DNA sequences at specified restriction enzyme sites.

Copyright © 2024 Samir Bertache

SPDX-License-Identifier: AGPL-3.0-or-later

===============================================================================

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import signal
import subprocess
import sys

# Setup logging
logging.basicConfig(level=logging.INFO)


def signal_handler(sig, frame, outF, outR):
    print(f"\nReceived signal {sig}. Terminating gracefully...")
    outF.terminate()  # Terminate the pigz processes
    outR.terminate()
    logging.info("\nProcess termination requested by signal")
    sys.exit(0)


def OpenOutput(TWrite, output_forward, output_reverse):
    # Open output files for writing
    outF = subprocess.Popen(
        ["pigz", "-c", "-p", str(TWrite)],
        stdin=subprocess.PIPE,
        stdout=open(output_forward, "wb"),
    )
    outR = subprocess.Popen(
        ["pigz", "-c", "-p", str(TWrite)],
        stdin=subprocess.PIPE,
        stdout=open(output_reverse, "wb"),
    )

    # Register signal handlers
    signal.signal(
        signal.SIGINT,
        lambda sig, frame: signal_handler(sig, frame, outF, outR),
    )  # Ctrl+C
    signal.signal(
        signal.SIGTSTP,
        lambda sig, frame: signal_handler(sig, frame, outF, outR),
    )  # Ctrl+Z

    return outF, outR


def ManagePigzProblems(outF, outR, output_forward, output_reverse):
    outF.stdin.close()
    outR.stdin.close()
    outF.wait()
    outR.wait()

    stdoutF, stderrF = outF.communicate()
    if stderrF:
        print(
            f"Error in pigz command for file {output_forward}: {stderrF}",
            flush=True,
        )

    stdoutR, stderrR = outR.communicate()
    if stderrR:
        print(
            f"Error in pigz command for file {output_reverse}: {stderrR}",
            flush=True,
        )


def write_pairs(
    Output_buffer,
    outF: subprocess.Popen,
    outR: subprocess.Popen,
    TFrag,
) -> None:
    finished_processes = 0
    while finished_processes < TFrag:
        try:
            data = Output_buffer.get()
            if data is None:
                finished_processes += 1
            else:
                outF.stdin.write("".join(data[0]).encode("utf-8"))
                outR.stdin.write("".join(data[1]).encode("utf-8"))

        except Exception as e:
            logging.error(f"Error in write_pairs: {e}")



"""
This script is a the samcut project, designed to process paired-end FASTQ files by fragmenting DNA sequences at specified restriction enzyme sites.

Copyright © 2024 Samir Bertache

SPDX-License-Identifier: AGPL-3.0-or-later

===============================================================================

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

from .Frag import process_items
from .Read import read_fastq_gzip_simultaneously_MyWay
from .Pretreatment import SearchInDataBase, Partitionning
from .WriteAndControl import OpenOutput, ManagePigzProblems, write_pairs



"""
This script is a the samcut project, designed to process paired-end FASTQ files by fragmenting DNA sequences at specified restriction enzyme sites.

Copyright © 2024 Samir Bertache

SPDX-License-Identifier: AGPL-3.0-or-later

===============================================================================

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import gzip
import os
import pytest
from parasplit.main import cut

@pytest.fixture(scope="module")
def setup_test_environment():
    # Ensure that the output directory for test exists
    os.makedirs("test/output_data", exist_ok=True)
    yield
    # Add any necessary post-test clean-up here


def check_file_content(output_file, expected_output_file):
    # Check that the content of the output file matches the expected result
    with gzip.open(output_file, "rt") as fA, gzip.open(expected_output_file, "rt") as fB:
        while True:
            nameA = fA.readline().rstrip()
            seqA = fA.readline().rstrip()
            fA.readline()  # Skip the '+'
            qualA = fA.readline().rstrip()

            nameB = fB.readline().rstrip()
            seqB = fB.readline().rstrip()
            fB.readline()  # Skip the '+'
            qualB = fB.readline().rstrip()
            
            assert (seqA == seqB) or (seqA[::-1] == seqB)
            
            if not seqA:
                break   
                
                
def test_process_file(setup_test_environment):
    # Paths to input and output files for the test
    input_file_forward = "test/input_data/R1.fq.gz"
    input_file_reverse = "test/input_data/R2.fq.gz"
    
    expected_output_file_forward_fr = "test/output_data/output_ref_R1.fq.gz"
    expected_output_file_reverse_fr = "test/output_data/output_ref_R2.fq.gz"  
    
    expected_output_file_forward_all = "test/output_data/output_ref_all_R1.fq.gz"
    expected_output_file_reverse_all = "test/output_data/output_ref_all_R2.fq.gz" 
    
    output_file_forward = "test/output_data/output_R1.fq.gz"
    output_file_reverse = "test/output_data/output_R2.fq.gz"

    # Execute your main function with the input files
    cut(
        input_file_forward,
        input_file_reverse,
        output_file_forward,
        output_file_reverse,
        ['DpnII'],
        0,
        "fr",
        4,
        False,
    )

    # Verify that the output file has been created
    assert os.path.exists(output_file_forward)
    assert os.path.exists(output_file_reverse)
    
    # Verify that the content of the output file matches the expected result in "fr" mode 
    check_file_content(output_file_reverse, expected_output_file_reverse_fr)
    check_file_content(output_file_forward, expected_output_file_forward_fr)
    
    
    # Execute your main function with the input files
    cut(
        input_file_forward,
        input_file_reverse,
        output_file_forward,
        output_file_reverse,
        ['DpnII'],
        0,
        "all",
        4,
        False,
    )
    
    # Verify that the content of the output file matches the expected result in "all" mode 
    check_file_content(output_file_reverse, expected_output_file_reverse_all)
    check_file_content(output_file_forward, expected_output_file_forward_all)

    # Clean up: remove the generated output files
    os.remove(output_file_reverse)
    os.remove(output_file_forward)




"""
This script is a the samcut project, designed to process paired-end FASTQ files by fragmenting DNA sequences at specified restriction enzyme sites.

Copyright © 2024 Samir Bertache

SPDX-License-Identifier: AGPL-3.0-or-later

===============================================================================

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import re
from typing import List, Tuple
import sys
from Bio.Restriction import RestrictionBatch
from Bio.Seq import Seq

logging.basicConfig(level=logging.INFO)


###############################  Common part #################################


def FindLigaSite(
    List_Enzyme: List[str], borderless: bool = False
) -> List[Tuple[re.Pattern, int]]:
    """
    This function finds the ligation sites for a given list of enzymes and
    their length.

    Parameters:
    List_Enzyme (List[str]): A list of enzymes for which to find the ligation
    sites.

    borderless (bool, optional): If True, the total length of the give and
                                        accept sites is used.
                                 If False, only the length of the give site
                                        is used. Default is False.

    Returns:
    List[Tuple[re.Pattern, int]]: A list of tuples, where each tuple contains
                                    a compiled regular expression
                                    pattern for the ligation site and the length
                                    of the site.
    """
    restriction_batch = RestrictionBatch(List_Enzyme)
    give_list = []
    accept_list = []
    ligation_site_list = []

    for enz in restriction_batch:
        site = enz.elucidate()
        fw_cut = site.find("^")
        rev_cut = site.find("_")

        # Purify give site
        give_site = site[:rev_cut].replace("^", "")
        while give_site[0] == "N":
            give_site = give_site[1:]
        give_list.append(give_site)

        # Purify accept site
        accept_site = site[fw_cut + 1:].replace("_", "")
        while accept_site[-1] == "N":
            accept_site = accept_site[:-1]
        accept_list.append(accept_site)

    # Find ligation site
    for give_site in give_list:
        for accept_site in accept_list:
            ligation_site = (give_site + accept_site).replace("N", ".")
            compiled_regex = re.compile(ligation_site)

            # Use total lenght for borderless
            if borderless:
                length = len(give_site) + len(accept_site)
            else:
                length = len(give_site)
            ligation_site_list.append((compiled_regex, length))

            # If ligation site is not palindromic
            reverse_complement_site = str(
                Seq(ligation_site).reverse_complement()
            )

            if ligation_site != reverse_complement_site:
                compiled_reverse_regex = re.compile(reverse_complement_site)
                # Use lenght of accept site for reverse complement site
                if borderless:
                    length = len(give_site) + len(accept_site)
                else:
                    length = len(accept_site)
                ligation_site_list.append((compiled_reverse_regex, length))

    return ligation_site_list


def SearchInDataBase(ListEnzyme, borderless=False):
    """
    _summary_ : Search enzyme in database and retrieve ligation site
    """
    if ListEnzyme == "No restriction enzyme found":
        print(ListEnzyme)
        sys.exit(0)
    else:
        if borderless:
            print("Mode Borderless")
        ligation_site_list = FindLigaSite(ListEnzyme.split(","), borderless)
        print(f"Ligation sites: {ligation_site_list}", flush=True)
        return ligation_site_list


def Partitionning(num_threads):
    TWrite = num_threads // 4
    TRead = num_threads // 8
    TFrag = num_threads - (TWrite * 2) - (TRead * 2)
    return TWrite, TRead, TFrag



"""
This script is a the samcut project, designed to process paired-end FASTQ files by fragmenting DNA sequences at specified restriction enzyme sites.

Copyright © 2024 Samir Bertache

SPDX-License-Identifier: AGPL-3.0-or-later

===============================================================================

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import re
from typing import Generator, List, Tuple

logging.basicConfig(level=logging.INFO)


#################### Specific part of Borderless option #####################


def find_positions_for_one_site_borderless(
    text: str, Enzyme: Tuple[re.Pattern, int]
) -> List[Tuple[int, int]]:
    """
    Find all positions of a specific pattern (RESite) in a given text using regular expressions.

    Parameters:
        text (str): The text to search in.
        Enzyme (Tuple[str, int]): The restriction enzyme site to search for, with the length of give_site or accept_site
        depend of the

    Examples:
        >>> find_positions_for_one_site_borderless("777777GA7TA7TC1717GA7TA7TC", (re.compile("GA.TA.TC"), 4))
        [[6, 4], [18, 4]]
    """
    regex, offset = Enzyme
    return [[m.start(), offset] for m in regex.finditer(text)]


def find_all_pos_borderless(
    text: str,
    ligation_site_list: List[Tuple[str, int]],
) -> List[int]:
    """
    Find all positions of a specific pattern (RESite) in a given text using regular expressions.

    Examples:
        >>> find_all_pos_borderless("777777GA7TA7TC1717GAATTCFFFFFF", [(re.compile("GAATTC"), 5), (re.compile("GA.TA.TC"), 4)])
        [[6, 4], [18, 5], [30, 0]]
    """
    AllSite = []
    for TupleRegex in ligation_site_list:
        AllSite += find_positions_for_one_site_borderless(text, TupleRegex)
    AllSite = AllSite + [[len(text), 0]]

    return sorted(AllSite, key=lambda x: x[0])


def IndexFragList_borderless(all_index_list, seed_size):
    """
    Find fragments position from a given list of indices

    If the fragment is shorter than the seed size, it is discarded
    If the fragment is longer than the seed size, it is added to the list

    The border corresponding to ligation site are discared

    Examples:
        >>> IndexFragList_borderless([[5, 4], [18, 5], [30, 0]], 6)
        [[9, 18], [23, 30]]
    """
    ListFragListIndex = []
    i = 0
    previous_position = 0

    # Skip border corresponding to ligation site in the list
    for i, value in enumerate(all_index_list):

        previous_border_len = all_index_list[i - 1][1]
        current_position = all_index_list[i][0]

        # Update value
        Start = int(previous_position + previous_border_len)
        End = current_position

        # Add fragment to the list if it is long enough
        if (End - Start) >= seed_size:
            ListFragListIndex.append([Start, End])

        # Update value :
        previous_position = current_position

    return ListFragListIndex


def index_list_single_borderless(
    sequence: str, ligation_site_list: List[Tuple[str, int]], seed_size
) -> List[List[int]]:
    """
    _summary_ : Find index position of fragments produce by digestion of
    sequences according to seed size

    :param sequence: _description_
    :param ligation_site_list: _description_
    :param seed_size: _description_
    :return: _description_
    """
    IndexPositions = find_all_pos_borderless(sequence, ligation_site_list)
    FragmentList = IndexFragList_borderless(IndexPositions, seed_size)
    return FragmentList


def index_list_borderless(
    Sequence: str,
    ligation_site_list,
    seed_size,
) -> List[str]:
    """
    _summary_ : Retrieve the index of the fragment sequence using the ligation
    sites and the seed size and without border (ligation site)

    :param seed_size: Minimal lenght of fragment
    :return: List of sequence fragment

    Examples:
        >>> index_list_borderless(["111111GAATTC222222", "111GAATTC222222"], [(re.compile("GAATTC"), 6)], 0)
        ([[0, 6], [12, 18]], [[0, 3], [9, 15]])
    """
    ListFragmentFor = index_list_single_borderless(
        Sequence[0], ligation_site_list, seed_size
    )
    ListFragmentRev = index_list_single_borderless(
        Sequence[1], ligation_site_list, seed_size
    )

    return ListFragmentFor, ListFragmentRev


############################# End Borderless ##################################


###############################  Classic part #################################


def find_positions_for_one_site(
    text: str, Enzyme: Tuple[str, int]
) -> List[int]:
    """
    Find all positions of a specific pattern (RESite) in a given text using regular expressions.

    Parameters:
        text (str): The text to search in.
        Enzyme (Tuple[str, int]): The restriction enzyme site to search for, with the length of give_site or accept_site
        depend of the

    Examples:
        >>> find_positions_for_one_site("AAGAATTCAA", (re.compile("GAATTC"), 5))
        [7]
        >>> find_positions_for_one_site("AAAAAGAATTCAAAAAGAATTCAAAAAGAATTC", (re.compile("GAATTC"), 5))
        [10, 21, 32]
        >>> find_positions_for_one_site("777777GA7TA7TC1717GA7TA7TC", (re.compile("GA.TA.TC"), 4))
        [10, 22]
    """
    return [
        int(match.start() + Enzyme[1])
        for match in re.finditer(Enzyme[0], text)
    ]


def find_all_pos(
    text: str, ligation_site_list: List[Tuple[re.Pattern, int]]
) -> List[int]:
    """
    Find all positions of a specific pattern (RESite) in a given text using regular expressions.

    Examples:
        >>> find_all_pos("777777GA7TA7TC1717GAATTCFFFFFF", [(re.compile("GAATTC"), 5), (re.compile("GA.TA.TC"), 4)])
        [[6, 4], [18, 5], [30, 0]]
    """
    AllSite = [0] + list(
        map(int, find_positions_for_one_site(text, ligation_site_list[0]))
    )
    if len(ligation_site_list) > 1:
        for TupleRegex in ligation_site_list[1:]:
            AllSite += list(
                map(int, find_positions_for_one_site(text, TupleRegex))
            )
    AllSite = [0] + AllSite + [len(text)]
    AllSite.sort()
    return AllSite


def IndexFragList(index_list, seed_size):
    """
    _summary_ : Find the index of the fragment which respect the minimum seed
    size

    """
    ListFragListIndex = []
    for i in range(len(index_list)):
        if i > 0:
            frag_size = index_list[i] - index_list[i - 1]
            if frag_size > seed_size:
                FragIndex = [index_list[i - 1], index_list[i]]
                ListFragListIndex.append(FragIndex)
    return ListFragListIndex


def index_list_single(
    sequence: str, ligation_site_list: List[Tuple[str, int]], seed_size
) -> List[List[int]]:
    """
    _summary_ : Find index position of fragments produce by digestion of
    sequences according to seed size
    """
    IndexPositions = find_all_pos(sequence, ligation_site_list)
    FragmentList = IndexFragList(IndexPositions, seed_size)
    return FragmentList


def index_list(
    Sequence: List[str],
    ligation_site_list: List[Tuple[re.Pattern, int]],
    seed_size,
) -> Tuple[List[List[int]]]:
    """
    _summary_ : Find the index of the fragment for forward and
    reverse sequences

    Examples:
        >>> index_list(["111111GAATTC222222", "111GAATTC222222"], [(re.compile("GAATTC"), 3)], 0)
        ([[0, 9], [9, 18]], [[0, 6], [6, 15]])
    """
    ListFragmentFor = index_list_single(
        Sequence[0], ligation_site_list, seed_size
    )
    ListFragmentRev = index_list_single(
        Sequence[1], ligation_site_list, seed_size
    )

    return ListFragmentFor, ListFragmentRev


################################ Mode All #####################################


def Create_Pairs_All(
    Sequence, seed_size, ligation_site_list, indexation
) -> Generator:
    """
    Create all possible pairs of fragments from given sequences.

    Parameters:
        Sequence (List[str]): List containing forward and reverse sequences.
        seed_size (int): Minimum size of the fragment to be considered.
        ligation_site_list (List[Tuple[re.Pattern, int]]): List of ligation sites with regex patterns and offsets.

    Yields:
        List: A list containing:
            - A unique identifier for the pair.
            - Information about the first fragment and which sequence it comes from (forward or reverse).
            - Information about the second fragment and which sequence it comes from (forward or reverse).
    """
    ListFragmentFor, ListFragmentRev = indexation(
        Sequence, ligation_site_list, seed_size
    )
    AllFrag = ListFragmentFor + ListFragmentRev
    NbFragFor = len(ListFragmentFor)
    for i, fragI in enumerate(AllFrag):
        for j, fragJ in enumerate(AllFrag):
            if i < j:
                if i < NbFragFor:
                    WhichFragFor = 0
                    SenseFor = 1
                else:
                    WhichFragFor = 1
                    SenseFor = -1
                if j < NbFragFor:
                    WhichFragRev = 0
                    SenseRev = -1
                else:
                    WhichFragRev = 1
                    SenseRev = 1

                Num = str(i) + str(j)
                if ((fragI[1] - fragI[0]) > seed_size) and (
                    (fragJ[1] - fragJ[0]) > seed_size
                ):
                    yield [
                        Num,
                        [fragI, WhichFragFor, SenseFor],
                        [fragJ, WhichFragRev, SenseRev],
                    ]


def processing_All(
    Name, Sequence, Quality, ligation_site_list, seed_size, indexation
):
    """
    Process the sequences to generate buffers for forward and reverse reads.

    Parameters:
        Name (str): Name of the read.
        Sequence (List[str]): List containing forward and reverse sequences.
        Quality (List[str]): List containing quality scores for forward and reverse sequences.
        seed_size (int): Minimum size of the fragment to be considered.
        ligation_site_list (List[Tuple[re.Pattern, int]]): List of ligation sites with regex patterns and offsets.

    Returns:
        Tuple[List[str], List[str]]: Buffers for forward and reverse reads with new names and sequences.
    """
    bufferF = []
    bufferR = []

    for BlocDePairs in Create_Pairs_All(
        Sequence, seed_size, ligation_site_list, indexation
    ):
        # Conservation of sense Forward and Reverse
        PremiereSeq = BlocDePairs[1][1]
        SenseFor = BlocDePairs[1][2]
        DeuxiemeSeq = BlocDePairs[2][1]
        SenseRev = BlocDePairs[2][2]

        # If you want understand this, see the comment after :
        NewName = str(Name.split(" ")[0]) + ":" + str(BlocDePairs[0])
        bufferF.append(
            f"{NewName}\n{Sequence[PremiereSeq][BlocDePairs[1][0][0]:BlocDePairs[1][0][1]][::SenseFor]}\n+\n{Quality[PremiereSeq][BlocDePairs[1][0][0]:BlocDePairs[1][0][1]][::SenseFor]}\n"
        )

        bufferR.append(
            f"{NewName}\n{Sequence[DeuxiemeSeq][BlocDePairs[2][0][0]:BlocDePairs[2][0][1]][::SenseRev]}\n+\n{Quality[DeuxiemeSeq][BlocDePairs[2][0][0]:BlocDePairs[2][0][1]][::SenseRev]}\n"
        )

    return bufferF, bufferR


"""
    For comprehensive explanation :

    BlocDePairs =   [Num,
                    [[StartIndex, EndIndex], CommingFromForwardOrReverse, ReverseOrNot],
                    [[StartIndex, EndIndex], CommingFCommingFromForwardOrReverseromReverse, ReverseOrNot]]


    bufferF = NewName
              Sequence[CommingFromForwardOrReverse][Index_Fragment][ReverseOrNot]
              +
              Qualitie[CommingFromForwardOrReverse][Index_Fragment][ReverseOrNot]
"""


################################ Mode FR #####################################


def Create_Pairs_Fr(
    fragments: List[List[int]],
) -> Generator[List[str | List[int]], None, None]:
    """
    Create pairs of forward and reverse fragments.

    Parameters:
        fragments (List[List[List[int]]]): List containing forward and reverse fragments. Each fragment is a list of start and end indices.

    Yields:
        List: A list containing:
            - A unique identifier for the pair.
            - The forward fragment indices.
            - The reverse fragment indices.

    """
    forward_fragments = fragments[0]
    reverse_fragments = fragments[1]

    for i, index_f_frag in enumerate(forward_fragments):
        for j, index_r_frag in enumerate(reverse_fragments):
            Num = str(i) + str(j)
            yield [Num, index_f_frag, index_r_frag]


def processing_Fr(
    TNom, TSeq, TQual, ligation_site_list, seed_size, indexation
) -> Tuple[List[str], List[str]]:
    """
    Process the sequences to generate buffers for forward and reverse reads based on fragment pairs.

    Parameters:
        item (Tuple[List[str], List[str], List[str]]): A tuple containing:
            - List of names.
            - List of sequences (forward and reverse).
            - List of quality scores (forward and reverse).
        ligation_site_list (List[Tuple[re.Pattern, int]]): List of ligation sites with regex patterns and offsets.
        seed_size (int): Minimum size of the fragment to be considered.

    Returns:
        Tuple[List[str], List[str]]: Buffers for forward and reverse reads with new names and sequences.
    """
    bufferF = []
    bufferR = []

    ListFrag = indexation(TSeq, ligation_site_list, seed_size)

    Currents_Pairs = Create_Pairs_Fr(ListFrag)

    for Current_Pair in Currents_Pairs:
        Num, index_f_frag, index_r_frag = Current_Pair

        NewName = str(TNom.split(" ")[0]) + ":" + str(Num)
        bufferF.append(
            f"{NewName}\n{TSeq[0][index_f_frag[0]:index_f_frag[1]]}\n+\n{TQual[0][index_f_frag[0]:index_f_frag[1]]}\n"
        )
        bufferR.append(
            f"{NewName}\n{TSeq[1][index_r_frag[0]:index_r_frag[1]]}\n+\n{TQual[1][index_r_frag[0]:index_r_frag[1]]}\n"
        )

    return bufferF, bufferR


################################ Common #####################################


def process_items(
    Input_Buffer,
    Output_buffer,
    ligation_site_list,
    seed_size,
    buffer_size,
    mode,
    borderless,
):
    """
    _summary_ : Process the sequences to generate FastQ sequences paired
    """
    BigBufferF = []
    BigBufferR = []

    if borderless:
        indexation = index_list_borderless
    else:
        indexation = index_list

    if mode == "all":
        Treatment = processing_All
    else:
        Treatment = processing_Fr

    while True:
        try:
            Items = Input_Buffer.get()
            if Items is None:
                break

            for item in Items:
                # Run the synchronous processing in a separate process
                bufferF, bufferR = Treatment(
                    item[0][0],
                    item[1],
                    item[2],
                    ligation_site_list,
                    seed_size,
                    indexation,
                )

                BigBufferF.extend(bufferF)
                BigBufferR.extend(bufferR)

                if len(BigBufferF) > buffer_size:
                    Output_buffer.put([BigBufferF, BigBufferR])
                    BigBufferF = []
                    BigBufferR = []

        except Exception as e:
            logging.error(f"Error in process items : {e}")

    if BigBufferF or BigBufferR:
        Output_buffer.put([BigBufferF, BigBufferR])
    Output_buffer.put(None)



"""
This script is a the samcut project, designed to process paired-end FASTQ files by fragmenting DNA sequences at specified restriction enzyme sites.

Copyright © 2024 Samir Bertache

SPDX-License-Identifier: AGPL-3.0-or-later

===============================================================================

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import subprocess

logging.basicConfig(level=logging.INFO)


def StopSignal(Queue, NumThreadFrag):
    """
    _summary_ : Add a stop signal to the queue for each thread
    """
    for _ in range(NumThreadFrag):
        Queue.put(None)


def read_fastq_gzip_simultaneously_MyWay(
    fileA: str, fileB: str, Queue, num_threads, NumThreadFrag
    ):
    """
    _summary_ : Read two fastq files simultaneously, decompress them with pigz,
    take a couple a read and put them into a queue by block
    """
    # Use pigz to decompress the input files
    procA = subprocess.Popen(
        ["pigz", "-dc", "-p", str(num_threads), fileA],
        stdout=subprocess.PIPE,
        text=True,
    )
    procB = subprocess.Popen(
        ["pigz", "-dc", "-p", str(num_threads), fileB],
        stdout=subprocess.PIPE,
        text=True,
    )

    Stacker = []
    try:
        while True:
            NomA = (procA.stdout.readline()).rstrip()
            seqA = (procA.stdout.readline()).rstrip()
            procA.stdout.readline()  # Skip +
            qualA = (procA.stdout.readline()).rstrip()

            NomB = (procB.stdout.readline()).rstrip()
            seqB = (procB.stdout.readline()).rstrip()
            procB.stdout.readline()  # Skip +
            qualB = (procB.stdout.readline()).rstrip()

            if not seqA or not seqB:
                break

            Stacker.append([[NomA, NomB], [seqA, seqB], [qualA, qualB]])

            if len(Stacker) > 10:
                Queue.put(Stacker)
                Stacker = []

        if len(Stacker) > 0:
            Queue.put(Stacker)
        Queue.put(None)

    except Exception as e:
        logging.error(f"Error in TakeOneItem: {e}")

    finally:
        StopSignal(Queue, NumThreadFrag)
        procA.wait()
        procB.wait()
